IMPLEMENTATION MODULE Conversions;

(**************************************************************************
   JAVA      Modula-2 utility for numeric conversions to/from strings.

   Copyright (c) 1998 by Mill Hill & Canterbury Corporation, Ltd.
**************************************************************************)

IMPORT SYSTEM, Strings;


PROCEDURE ShortIntToStr
( i        : SHORTINT;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
BEGIN
  LongIntToStr( i, s, done );
END ShortIntToStr;


PROCEDURE IntToStr
( i        : INTEGER;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
BEGIN
  LongIntToStr( i, s, done );
END IntToStr;


PROCEDURE LongIntToStr
( i        : LONGINT;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
VAR
  j,k    : LONGCARD;
  n      : LONGINT;
BEGIN
  k := HIGH( s );
  j := 0;
  IF k = 0 THEN
    done := FALSE; RETURN;
  END;
  IF i = 0 THEN
    s[ 0 ] := '0';
    s[ 1 ] := 0C;
    done := TRUE; RETURN;
  END;
  IF i = MIN( LONGINT ) THEN
    IF k >= 11 THEN
      Strings.Copy( "-2147483648", 0, 11, s );
      done := TRUE; RETURN;
    END;
    done := FALSE; RETURN;
  END;
  IF i < 0 THEN
    s[ 0 ] := '-';
    j := 1;
    i := -i;
  END;
  s[ k ] := 0C;
  WHILE (k > j) AND (i > 0) DO
    DEC( k );
    n := i DIV 10;
    s[ k ] := CHR( ORD( '0' ) + ORD( i - n * 10 ) );
    i := n;
  END;
  IF i > 0 THEN
    done := FALSE; RETURN;
  END;
  IF k > j THEN
    WHILE s[ k ] <> 0C DO
      s[ j ] := s[ k ];
      INC( j );
      INC( k );
    END;
    s[ j ] := 0C;
  END;
  done := TRUE;
END LongIntToStr;


PROCEDURE StrToShortInt
( s        : ARRAY OF CHAR;
  VAR i    : SHORTINT;
  VAR done : BOOLEAN
);
VAR
  j      : LONGINT;
BEGIN
  StrToLongInt( s, j, done );
  IF done THEN
    IF j >= MIN( SHORTINT ) THEN
      IF j <= MAX( SHORTINT ) THEN
        i := SHORT( SHORT( j ) );
        done := TRUE; RETURN;
      END;
    END;
  END;
  done := FALSE;
END StrToShortInt;


PROCEDURE StrToInt
( s        : ARRAY OF CHAR;
  VAR i    : INTEGER;
  VAR done : BOOLEAN
);
VAR
  j      : LONGINT;
BEGIN
  StrToLongInt( s, j, done );
  IF done THEN
    IF j >= MIN( INTEGER ) THEN
      IF j <= MAX( INTEGER ) THEN
        i := SHORT( j );
        done := TRUE; RETURN;
      END;
    END;
  END;
  done := FALSE;
END StrToInt;


PROCEDURE StrToLongInt
( s        : ARRAY OF CHAR;
  VAR i    : LONGINT;
  VAR done : BOOLEAN
);
VAR
  j,k,m    : LONGCARD;
  neg      : BOOLEAN;
  overflow : BOOLEAN;
BEGIN
  k := Strings.Length( s );
  neg := FALSE;
  overflow := FALSE;
  j := 0;
  i := 0;
  WHILE (j < k) AND (s[ j ] = ' ') DO
    INC( j );
  END;
  IF s[ j ] = '-' THEN
    neg := TRUE;
    INC( j );
  ELSIF s[ j ] = '+' THEN
    INC( j );
  END;
  m := j;
  WHILE (j < k) AND (s[ j ] = '0') DO
    INC( j );
  END;
  WHILE (j < k) AND ('0' <= s[ j ]) AND (s[ j ] <= '9') DO
    i := i * 10 + VAL( LONGINT, ORD( s[ j ] )-ORD( '0' ) );
    IF i < 0 THEN
      IF neg THEN
        IF i > MIN( LONGINT ) THEN
          overflow := TRUE;
        END
      ELSE
        overflow := TRUE;
      END;
    END;
    INC( j );
  END;
  IF neg THEN
    i := -i;
  END;
  done := (j > m) AND (j >= k) AND NOT overflow;
END StrToLongInt;


PROCEDURE ShortCardToStr
( c        : SHORTCARD;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
BEGIN
  LongCardToStr( c, s, done );
END ShortCardToStr;


PROCEDURE CardToStr
( c        : CARDINAL;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
BEGIN
  LongCardToStr( c, s, done );
END CardToStr;


PROCEDURE LongCardToStr
( c        : LONGCARD;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
VAR
  j,k,n  : LONGCARD;
BEGIN
  k := HIGH( s );
  IF k = 0 THEN
    done := FALSE; RETURN;
  END;
  IF c = 0 THEN
    s[ 0 ] := '0';
    s[ 1 ] := 0C;
    done := TRUE; RETURN;
  END;
  s[ k ] := 0C;
  WHILE (k > 0) AND (c > 0) DO
    DEC( k );
    n := c DIV 10;
    s[ k ] := CHR( ORD( '0' ) + ORD( c - n * 10 ) );
    c := n;
  END;
  IF c > 0 THEN
    done := FALSE; RETURN;
  END;
  IF k > 0 THEN
    j := 0;
    WHILE s[ k ] <> 0C DO
      s[ j ] := s[ k ];
      INC( j );
      INC( k );
    END;
    s[ j ] := 0C;
  END;
  done := TRUE;
END LongCardToStr;


PROCEDURE StrToShortCard
( s        : ARRAY OF CHAR;
  VAR c    : SHORTCARD;
  VAR done : BOOLEAN
);
VAR
  j      : LONGCARD;
BEGIN
  StrToLongCard( s, j, done );
  IF done THEN
    IF j <= MAX( SHORTCARD ) THEN
      c := SHORT( SHORT( j ) );
      RETURN;
    END;
  END;
  done := FALSE;
END StrToShortCard;


PROCEDURE StrToCard
( s        : ARRAY OF CHAR;
  VAR c    : CARDINAL;
  VAR done : BOOLEAN
);
VAR
  j      : LONGCARD;
BEGIN
  StrToLongCard( s, j, done );
  IF done THEN
    IF j <= MAX( CARDINAL ) THEN
      c := SHORT( j );
      RETURN;
    END;
  END;
  done := FALSE; RETURN;
END StrToCard;


PROCEDURE StrToLongCard
( s        : ARRAY OF CHAR;
  VAR c    : LONGCARD;
  VAR done : BOOLEAN
);
VAR
  j,k,m,n  : LONGCARD;
  overflow : BOOLEAN;
BEGIN
  k := Strings.Length( s );
  j := 0;
  c := 0;
  overflow := FALSE;
  WHILE (j < k) AND (s[ j ] = ' ') DO
    INC( j );
  END;
  IF s[ j ] = '+' THEN
    INC( j );
  END;
  m := j;
  WHILE (j < k) AND (s[ j ] = '0') DO
    INC( j );
  END;
  WHILE (j < k) AND ('0' <= s[ j ]) AND (s[ j ] <= '9') DO
    n := c;
    overflow := overflow OR (n > MAX(LONGCARD) DIV 10);
    c := c * 10 + VAL( LONGCARD, ORD( s[ j ] )-ORD( '0' ) );
    overflow := overflow OR (c < n);
    INC( j );
  END;
  done := (j >= k) AND (j > m) AND NOT overflow;
END StrToLongCard;


PROCEDURE ShortHexToStr
( h        : SYSTEM.BYTE;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
BEGIN
  LongHexToStr( LONG( LONG( VAL( SHORTCARD, ORD(h) ) ) ), s, done );
END ShortHexToStr;


PROCEDURE HexToStr
( h        : SYSTEM.WORD;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
BEGIN
  LongHexToStr( LONG( VAL( CARDINAL, ORD(h) ) ), s, done );
END HexToStr;


PROCEDURE LongHexToStr
( h        : SYSTEM.LONGWORD;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
VAR
  j,k,c  : LONGCARD;
  m,n    : LONGCARD;
BEGIN
  c := h;
  k := HIGH( s );
  IF k < 2 THEN
    done := FALSE; RETURN;
  END;
  IF c = 0 THEN
    s[ 0 ] := '0';
    s[ 1 ] := 'H';
    s[ 2 ] := 0C;
    done := TRUE; RETURN;
  END;
  s[ k ] := 0C;
  DEC( k );
  s[ k ] := 'H';
  WHILE (k > 0) AND (c > 0) DO
    DEC( k );
    n := c DIV 16;
    m := ORD( c - n * 16 );
    IF m < 10 THEN
      s[ k ] := CHR( ORD( '0' ) + m );
    ELSE
      s[ k ] := CHR( ORD( 'A' ) + m - 10 );
    END;
    c := n;
  END;
  IF c > 0 THEN
    done := FALSE; RETURN;
  END;
  IF (s[ k ] < '0') OR (s[ k ] > '9') THEN
    IF k > 0 THEN
      DEC( k );
      s[ k ] := '0';
    ELSE
      done := FALSE; RETURN;
    END;
  END;
  IF k > 0 THEN
    j := 0;
    WHILE s[ k ] <> 0C DO
      s[ j ] := s[ k ];
      INC( j );
      INC( k );
    END;
    s[ j ] := 0C;
  END;
  done := TRUE;
END LongHexToStr;


PROCEDURE StrToShortHex
( s        : ARRAY OF CHAR;
  VAR h    : SYSTEM.BYTE;
  VAR done : BOOLEAN
);
VAR
  j      : LONGCARD;
BEGIN
  StrToLongHex( s, j, done );
  IF done THEN
    IF j <= MAX( SHORTCARD ) THEN
      h := SHORT( SHORT( j ) );
      RETURN;
    END;
  END;
  done := FALSE;
END StrToShortHex;


PROCEDURE StrToHex
( s        : ARRAY OF CHAR;
  VAR h    : SYSTEM.WORD;
  VAR done : BOOLEAN
);
VAR
  j      : LONGCARD;
BEGIN
  StrToLongHex( s, j, done );
  IF done THEN
    IF j <= MAX( CARDINAL ) THEN
      h := SHORT( j );
      RETURN;
    END;
  END;
  done := FALSE;
END StrToHex;


PROCEDURE StrToLongHex
( s        : ARRAY OF CHAR;
  VAR h    : SYSTEM.LONGWORD;
  VAR done : BOOLEAN
);
VAR
  j,k,c,m,n  : LONGCARD;
  overflow   : BOOLEAN;
BEGIN
  k := Strings.Length( s );
  j := 0;
  c := 0;
  overflow := FALSE;
  WHILE (j < k) AND (s[ j ] = ' ') DO
    INC( j );
  END;
  IF s[ j ] = '+' THEN
    INC( j );
  END;
  IF (j < k) AND NOT (('0' <= s[ j ]) AND (s[ j ] <= '9')) THEN
    done := FALSE; RETURN;
  END;
  m := j;
  WHILE (j < k) AND (s[ j ] = '0') DO
    INC( j );
  END;
  WHILE (j < k) AND
        ( ('0' <= s[ j ]) AND (s[ j ] <= '9') OR
          ('A' <= s[ j ]) AND (s[ j ] <= 'F')
        ) DO
    n := c;
    overflow := overflow OR (c > MAX(LONGCARD) DIV 16);
    c := c * 16;
    IF ('0' <= s[ j ]) AND (s[ j ] <= '9') THEN
      c := c + ORD( s[ j ] ) - ORD( '0' );
    ELSE
      c := c + ORD( s[ j ] ) - ORD( 'A' ) + 10;
    END;
    overflow := overflow OR (c < n);
    INC( j );
  END;
  h := c;
  IF (j < k) AND (s[ j ] = 'H') THEN
    INC( j );
  ELSE
    done := FALSE; RETURN;
  END;
  done := (j >= k) AND (j > m) AND NOT overflow;
END StrToLongHex;


PROCEDURE ShortOctToStr
( o        : SYSTEM.BYTE;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
BEGIN
  LongOctToStr( LONG( LONG( VAL( SHORTCARD, ORD(o) ) ) ), s, done );
END ShortOctToStr;


PROCEDURE OctToStr
( o        : SYSTEM.WORD;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
BEGIN
  LongOctToStr( LONG( VAL( CARDINAL, ORD(o) ) ), s, done );
END OctToStr;


PROCEDURE LongOctToStr
( o        : SYSTEM.LONGWORD;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
VAR
  j,k,c  : LONGCARD;
  n      : LONGCARD;
BEGIN
  c := o;
  k := HIGH( s );
  IF k < 2 THEN
    done := FALSE; RETURN;
  END;
  IF c = 0 THEN
    s[ 0 ] := '0';
    s[ 1 ] := 'B';
    s[ 2 ] := 0C;
    done := TRUE; RETURN;
  END;
  s[ k ] := 0C;
  DEC( k );
  s[ k ] := 'B';
  WHILE (k > 0) AND (c > 0) DO
    DEC( k );
    n := c DIV 8;
    s[ k ] := CHR( ORD( '0' ) + ORD( c - n * 8 ) );
    c := n;
  END;
  IF c > 0 THEN
    done := FALSE; RETURN;
  END;
  IF k > 0 THEN
    j := 0;
    WHILE s[ k ] <> 0C DO
      s[ j ] := s[ k ];
      INC( j );
      INC( k );
    END;
    s[ j ] := 0C;
  END;
  done := TRUE;
END LongOctToStr;


PROCEDURE StrToShortOct
( s        : ARRAY OF CHAR;
  VAR o    : SYSTEM.BYTE;
  VAR done : BOOLEAN
);
VAR
  j      : LONGCARD;
BEGIN
  StrToLongOct( s, j, done );
  IF done THEN
    IF j <= MAX( SHORTCARD ) THEN
      o := SHORT( SHORT( j ) );
      RETURN;
    END;
  END;
  done := FALSE;
END StrToShortOct;


PROCEDURE StrToOct
( s        : ARRAY OF CHAR;
  VAR o    : SYSTEM.WORD;
  VAR done : BOOLEAN
);
VAR
  j      : LONGCARD;
BEGIN
  StrToLongOct( s, j, done );
  IF done THEN
    IF j <= MAX( CARDINAL ) THEN
      o := SHORT( j );
      RETURN;
    END;
  END;
  done := FALSE;
END StrToOct;


PROCEDURE StrToLongOct
( s        : ARRAY OF CHAR;
  VAR o    : SYSTEM.LONGWORD;
  VAR done : BOOLEAN
);
VAR
  j,k,c,m,n  : LONGCARD;
  overflow   : BOOLEAN;
BEGIN
  k := Strings.Length( s );
  j := 0;
  c := 0;
  overflow := FALSE;
  WHILE (j < k) AND (s[ j ] = ' ') DO
    INC( j );
  END;
  IF s[ j ] = '+' THEN
    INC( j );
  END;
  IF (j < k) AND NOT (('0' <= s[ j ]) AND (s[ j ] <= '7')) THEN
    done := FALSE; RETURN;
  END;
  m := j;
  WHILE (j < k) AND (s[ j ] = '0') DO
    INC( j );
  END;
  WHILE (j < k) AND ('0' <= s[ j ]) AND (s[ j ] <= '7') DO
    n := c;
    overflow := overflow OR (c > MAX(LONGCARD) DIV 8);
    c := c * 8 + ORD( s[ j ] ) - ORD( '0' );
    overflow := overflow OR (c < n);
    INC( j );
  END;
  o := c;
  IF (j < k) AND (s[ j ] = 'B') THEN
    INC( j );
  ELSE
    done := FALSE; RETURN;
  END;
  done := (j >= k) AND (j > m) AND NOT overflow;
END StrToLongOct;


END Conversions.
