IMPLEMENTATION MODULE FileSystem;

(*************************************************************************
   JAVA      Modula-2 standard utility for file system management.

   Copyright (c) 1998 by Mill Hill & Canterbury Corporation, Ltd.
*************************************************************************)


IMPORT SYSTEM;
IMPORT Strings;

VAR
  tempcount : LONGINT;

VAR
  Done            : LONGCARD;
  NotDone         : LONGCARD;
  LockError       : LONGCARD;
  PermissionError : LONGCARD;
  CallError       : LONGCARD;
  UnknownMedium   : LONGCARD;
  UnknownFile     : LONGCARD;
  FilenameError   : LONGCARD;
  TooManyFiles    : LONGCARD;
  MediumFull      : LONGCARD;
  ParamError      : LONGCARD;



PROCEDURE NewTempName( VAR FileName : ARRAY OF CHAR );
BEGIN
  SYSTEM.INLINE
  (
    java.io.File     temp;
    java.lang.String tempname;
    do
    {
      tempcount += 1;
      tempname = "$"+tempcount+".tmp";
      temp = new java.io.File( tempname );
    } while (temp.exists());
    int j = FileName.length;
    int i = 0;
    int k = tempname.length();
    while ((i < j-1) && (i < k))
    {
      FileName[i] = tempname.charAt(i);
      i += 1;
    }
    FileName[i] = (char)0;
  );
END NewTempName;


PROCEDURE Create( VAR f : File; FileName : ARRAY OF CHAR );
BEGIN
  f.eof := TRUE;
  Strings.Assign( FileName, f.name );
  IF Strings.Length( f.name ) > 0 THEN
    (* open a new permanent file *)
    f.tmp := FALSE;
    SYSTEM.INLINE
    (
      java.lang.String name = new java.lang.String( f.name );
      try
      {
        f.res = Done;
        java.io.FileOutputStream s = new java.io.FileOutputStream( name );
        s.close();
      }
      catch (java.io.IOException e)
      {
        f.res = NotDone;
      }
      catch (java.lang.SecurityException e)
      {
        f.res = PermissionError;
      }
      if (f.res == Done)
      {
        try
        {
          f.id = new java.io.RandomAccessFile( name, "rw" );
        }
        catch (java.io.IOException e)
        {
          f.res = NotDone;
        }
        catch (java.lang.SecurityException e)
        {
          f.res = PermissionError;
        }
      }
    );
  ELSE
    (* open a temporary file *)
    NewTempName( f.name );
    IF Strings.Length( f.name ) > 0 THEN
      Create( f, f.name );
      f.tmp := TRUE;
    ELSE
      f.res := filenameerror;
    END;
  END;
END Create;


PROCEDURE Lookup( VAR f : File; FileName : ARRAY OF CHAR; isNew : BOOLEAN );
BEGIN
  f.tmp := FALSE;
  f.name[0] := 0C;
  f.res := done;
  SYSTEM.INLINE
  (
    java.io.File handle = new java.io.File( new java.lang.String( FileName ) );
    try
    {
      if ((!handle.exists()) || (!handle.isFile()))
      {
        handle = null;
        f.res = UnknownFile;
      }
    }
    catch ( java.lang.SecurityException e )
    {
      handle = null;
      f.res = PermissionError;
    };
    if (handle != null)
    {
      try
      {
        f.id = new java.io.RandomAccessFile( handle, "rw" );
      }
      catch ( java.io.IOException e )
      {
        f.res = NotDone;
        handle = null;
      }
      catch ( java.lang.SecurityException e )
      {
        f.res = PermissionError;
        handle = null;
      };
    }
  );
  IF f.res = done THEN
    f.eof := FALSE;
    Strings.Assign( FileName, f.name );
  ELSIF NOT isNew THEN
    f.eof := FALSE;
  ELSE
    Create( f, FileName );
  END;
END Lookup;


PROCEDURE Close( VAR f : File );
BEGIN
  Doio( f );
  SYSTEM.INLINE
  (
    f.res = Done;
    if (f.id instanceof java.io.RandomAccessFile)
    {
      try
      {
        ((java.io.RandomAccessFile)f.id).close();
      }
      catch ( java.lang.SecurityException e )
      {
        f.res = PermissionError;
      }
      catch ( java.io.IOException e )
      {
        f.res = NotDone;
      };
      f.id = null;
    }
    else
    {
      f.res = ParamError;
    }
  );
  IF f.res = done THEN
    f.eof := FALSE;
    IF f.tmp THEN
      Delete( f );
    END;
  END;
END Close;


PROCEDURE Delete( VAR f : File );
BEGIN
  SYSTEM.INLINE
  (
    if (!(f.id instanceof java.io.RandomAccessFile))
    {
      f.res = ParamError;
    }
    else
    {
      try
      {
        ((java.io.RandomAccessFile)f.id).close();
      }
      catch (java.io.IOException e)
      {
      }
      f.id = null;
      java.io.File handle = new java.io.File(new String( f.name ));
      f.res = NotDone;
      try
      {
        if (handle.delete())
        {
          f.res = Done;
        }
      }
      catch (java.lang.SecurityException e)
      {
        f.res = PermissionError;
      }
    }
  );
  IF f.res = done THEN
    f.name[ 0 ] := 0C;
    f.tmp := FALSE;
    f.eof := FALSE;
  END;
END Delete;


PROCEDURE Rename( VAR f : File; FileName : ARRAY OF CHAR );
VAR
  i             : LONGINT;
  OldName       : ARRAY [0..127] OF CHAR;
  OldPos        : LONGCARD;
  SavedResult   : LONGINT;
  OldTmp        : BOOLEAN;
  OldEof        : BOOLEAN;
BEGIN
  f.res := done;
  SYSTEM.INLINE
  (
    if (!(f.id instanceof java.io.RandomAccessFile))
    {
      f.res = ParamError;
    }
  );
  IF f.res <> done THEN
    RETURN;
  END;
  i := Strings.Length( FileName );
  IF (i = 0) AND f.tmp THEN
    RETURN;
  END;
  Strings.Assign( f.name, OldName );
  GetLongPos( f, OldPos );
  IF f.res <> done THEN
    RETURN;
  END;
  OldTmp := f.tmp;
  OldEof := f.eof;
  f.tmp := FALSE;
  Close( f );
  IF f.res <> done THEN
    f.tmp := OldTmp;
    f.eof := OldEof;
    RETURN;
  END;
  IF i > 0 THEN
    SYSTEM.INLINE
    (
      try
      {
        java.io.File oldFile = new java.io.File(new String( OldName ));
        java.io.File newFile = new java.io.File(new String( FileName ));
        if (oldFile.renameTo(newFile))
        {
          f.res = Done;
        }
        else
        {
          f.res = NotDone;
        }
      }
      catch (java.lang.SecurityException e)
      {
        f.res = PermissionError;
      }
    );
  ELSE
    NewTempName( f.name );
    IF Strings.Length( f.name ) > 0 THEN
      SYSTEM.INLINE
      (
        try
        {
          java.io.File oldFile = new java.io.File(new String( OldName ));
          java.io.File newFile = new java.io.File(new String( f.name ));
          if (oldFile.renameTo(newFile))
          {
            f.res = Done;
          }
          else
          {
            f.res = NotDone;
          }
        }
        catch (java.lang.SecurityException e)
        {
          f.res = PermissionError;
        }
      );
    ELSE
      f.res := notdone;
    END;
  END;
  IF f.res <> done THEN
    (* Rename has not been successful, reopen with old name *)
    SavedResult := f.res;
    Lookup( f, OldName, FALSE );
    IF f.res = done THEN
      f.tmp := OldTmp;
      SetLongPos( f, OldPos );
      IF f.res = done THEN
        f.eof := OldEof;
      END;
    END;
    f.res := SavedResult;
  ELSE
    (* Rename has been successful, reopen with new name *)
    Lookup( f, FileName, FALSE );
    IF f.res = done THEN
      SetLongPos( f, OldPos );
      IF f.res = done THEN
        f.eof := OldEof;
      END;
    END;
  END;
END Rename;



PROCEDURE SetRead( VAR f : File );
BEGIN
  SYSTEM.INLINE
  (
    if (f.id instanceof java.io.OutputStreamWriter)
    {
      f.res = ParamError;
    }
    else if (f.id instanceof java.io.InputStreamReader)
    {
      f.res = Done;
    }
    else if (f.id instanceof java.io.RandomAccessFile)
    {
      f.res = Done;
    }
    else
    {
      f.res = NotDone;
    }
  );
END SetRead;


PROCEDURE SetWrite( VAR f : File );
BEGIN
  SYSTEM.INLINE
  (
    if (f.id instanceof java.io.OutputStreamWriter)
    {
      f.res = Done;
    }
    else if (f.id instanceof java.io.InputStreamReader)
    {
      f.res = ParamError;
    }
    else if (f.id instanceof java.io.RandomAccessFile)
    {
      java.io.RandomAccessFile id = null;
      try
      {
        java.io.File handle = new java.io.File( new java.lang.String( f.name ) );
        id = new java.io.RandomAccessFile( handle, "rw" );
        f.res = Done;
      }
      catch (java.io.IOException e)
      {
        f.res = NotDone;
      }
      catch (java.lang.SecurityException e)
      {
        f.res = PermissionError;
      }
      if (id != null)
      {
        f.id = id;
      }
    }
    else
    {
      f.res = NotDone;
    }
  );
END SetWrite;


PROCEDURE SetModify( VAR f : File );
BEGIN
  SYSTEM.INLINE
  (
    if (f.id == null)
    {
      f.res = NotDone;
    }
    else if (f.id instanceof java.io.RandomAccessFile)
    {
      SetWrite( f );
    }
    else
    {
      f.res = ParamError;
    }
  );
END SetModify;


PROCEDURE SetOpen( VAR f : File );
BEGIN
  Doio( f );
  f.res := done;
END SetOpen;


PROCEDURE Doio( VAR f : File );
BEGIN
  SYSTEM.INLINE
  (
    if (f.id == null)
    {
      f.res = NotDone;
    }
    else if (f.id instanceof java.io.OutputStreamWriter)
    {
      try
      {
        ((java.io.OutputStreamWriter)f.id).flush();
        f.res = Done;
      }
      catch (java.io.IOException e)
      {
        f.res = NotDone;
      }
    }
    else if (f.id instanceof java.io.RandomAccessFile)
    {
      f.res = Done;
    }
    else if (f.id instanceof java.io.InputStreamReader)
    {
      f.res = Done;
    }
    else
    {
      f.res = ParamError;
    }
  );
END Doio;


PROCEDURE SetPos( VAR f : File ; HighPos, LowPos : CARDINAL );
VAR
  Pos     : LONGCARD;
BEGIN
  Pos := LONG( HighPos ) * (LONG( MAX( CARDINAL ) )+1) + LONG( LowPos );
  SetLongPos( f, Pos );
END SetPos;


PROCEDURE GetPos( VAR f : File; VAR HighPos, LowPos : CARDINAL );
VAR
  i : LONGCARD;
BEGIN
  GetLongPos( f, i );
  HighPos := SHORT( i DIV ( LONG( MAX( CARDINAL ) )+1 ) );
  LowPos := SHORT( i );
END GetPos;


PROCEDURE Length( VAR f : File; VAR HighPos, LowPos : CARDINAL );
VAR
  i : LONGCARD;
BEGIN
  LongLength( f, i );
  HighPos := SHORT( i DIV LONG( MAX( CARDINAL ) ) + 1 );
  LowPos := SHORT( i  );
END Length;


PROCEDURE SetLongPos( VAR f : File; Pos : LONGCARD );
BEGIN
  SYSTEM.INLINE
  (
    if (f.id instanceof java.io.RandomAccessFile)
    {
      try
      {
        ((java.io.RandomAccessFile)f.id).seek( Pos );
        f.res = Done;
      }
      catch (java.io.IOException e)
      {
        f.res = NotDone;
      }
    }
    else
    {
      f.res = ParamError;
    }
  );
END SetLongPos;


PROCEDURE GetLongPos( VAR f : File; VAR Pos : LONGCARD );
BEGIN
  SYSTEM.INLINE
  (
    if (f.id instanceof java.io.RandomAccessFile)
    {
      try
      {
        Pos.val = (int)((java.io.RandomAccessFile)f.id).getFilePointer();
        f.res = Done;
      }
      catch (java.io.IOException e)
      {
        f.res = NotDone;
      }
    }
    else
    {
      f.res = ParamError;
    }
  );
END GetLongPos;


PROCEDURE LongLength( VAR f : File; VAR LongPos : LONGCARD );
BEGIN
  SYSTEM.INLINE
  (
    if (f.id instanceof java.io.RandomAccessFile)
    {
      try
      {
        LongPos.val = (int)((java.io.RandomAccessFile)f.id).length();
        f.res = Done;
      }
      catch (java.io.IOException e)
      {
        f.res = NotDone;
      }
    }
    else
    {
      f.res = ParamError;
    }
  );
END LongLength;


PROCEDURE Reset( VAR f : File );
BEGIN
  SetOpen( f );
  IF f.res = done THEN
    SetLongPos( f, 0 );
  END;
END Reset;


PROCEDURE Again( VAR f : File );
VAR
  Pos : LONGCARD;
BEGIN
  GetLongPos( f, Pos );
  IF (f.res = done) AND (Pos > 0) THEN
    SetLongPos( f, Pos-1 );
  END;
END Again;


PROCEDURE ReadByte( VAR f : File; VAR b : SYSTEM.BYTE );
BEGIN
  SYSTEM.INLINE
  (
    b.val = (byte)0;
    if (f.id instanceof java.io.InputStreamReader)
    {
      int val = 0;
      try
      {
        java.io.InputStreamReader id = (java.io.InputStreamReader)f.id;
        val = id.read();
        f.res = Done;
      }
      catch (java.io.IOException e)
      {
        val = 0;
        f.res = NotDone;
      }
      b.val = (byte)val;
      if (val == -1)
      {
        b.val = (byte)0;
        f.res = Done;
        f.eof = true;
      }
    }
    else if (f.id instanceof java.io.OutputStreamWriter)
    {
      f.res = ParamError;
    }
    else if (f.id instanceof java.io.RandomAccessFile)
    {
      try
      {
        b.val = ((java.io.RandomAccessFile)f.id).readByte();
        f.res = Done;
      }
      catch (java.io.EOFException e)
      {
        f.eof = true;
        f.res = Done;
        b.val = (byte)0;
      }
      catch (java.io.IOException e)
      {
        f.res = NotDone;
      }
    }
    else
    {
      f.res = ParamError;
    }
  );
END ReadByte;


PROCEDURE ReadWord( VAR f : File; VAR w : SYSTEM.WORD );
BEGIN
  SYSTEM.INLINE
  (
    w.val = (short)0;
    if (f.id instanceof java.io.InputStreamReader)
    {
      f.res = ParamError;
    }
    else if (f.id instanceof java.io.OutputStreamWriter)
    {
      f.res = ParamError;
    }
    else if (f.id instanceof java.io.RandomAccessFile)
    {
      try
      {
        w.val = ((java.io.RandomAccessFile)f.id).readShort();
        f.res = Done;
      }
      catch (java.io.EOFException e)
      {
        f.eof = true;
        f.res = Done;
        w.val = (short)0;
      }
      catch (java.io.IOException e)
      {
        f.res = NotDone;
      }
    }
    else
    {
      f.res = ParamError;
    }
  );
END ReadWord;


PROCEDURE ReadLongWord( VAR f : File; VAR lw : SYSTEM.LONGWORD );
BEGIN
  SYSTEM.INLINE
  (
    lw.val = 0;
    if (f.id instanceof java.io.InputStreamReader)
    {
      f.res = ParamError;
    }
    else if (f.id instanceof java.io.OutputStreamWriter)
    {
      f.res = ParamError;
    }
    else if (f.id instanceof java.io.RandomAccessFile)
    {
      try
      {
        lw.val = ((java.io.RandomAccessFile)f.id).readInt();
        f.res = Done;
      }
      catch (java.io.EOFException e)
      {
        f.eof = true;
        f.res = Done;
        lw.val = 0;
      }
      catch (java.io.IOException e)
      {
        f.res = NotDone;
      }
    }
    else
    {
      f.res = ParamError;
    }
  );
END ReadLongWord;


PROCEDURE ReadChar( VAR f : File; VAR ch : CHAR );
BEGIN
  SYSTEM.INLINE
  (
    ch.val = (char)0;
    if (f.id instanceof java.io.InputStreamReader)
    {
      int val = 0;
      try
      {
        java.io.InputStreamReader id = (java.io.InputStreamReader)f.id;
        val = id.read();
        f.res = Done;
      }
      catch (java.io.IOException e)
      {
        val = 0;
        f.res = NotDone;
      }
      ch.val = (char)val;
      if (val == -1)
      {
        ch.val = (char)0;
        f.res = Done;
        f.eof = true;
      }
    }
    else if (f.id instanceof java.io.OutputStreamWriter)
    {
      f.res = ParamError;
    }
    else if (f.id instanceof java.io.RandomAccessFile)
    {
      try
      {
        ch.val = ((java.io.RandomAccessFile)f.id).readChar();
        f.res = Done;
      }
      catch (java.io.EOFException e)
      {
        f.eof = true;
        f.res = Done;
        ch.val = 0;
      }
      catch (java.io.IOException e)
      {
        f.res = NotDone;
      }
    }
    else
    {
      f.res = ParamError;
    }
  );
END ReadChar;



PROCEDURE WriteByte( VAR f : File; b : SYSTEM.BYTE );
BEGIN
  SYSTEM.INLINE
  (
    if (f.id instanceof java.io.InputStreamReader)
    {
      f.res = ParamError;
    }
    else if (f.id instanceof java.io.OutputStreamWriter)
    {
      try
      {
        ((java.io.OutputStreamWriter)f.id).write( b & 0xFF );
        f.res = Done;
      }
      catch (java.io.IOException e)
      {
        f.res = NotDone;
      }
    }
    else if (f.id instanceof java.io.RandomAccessFile)
    {
      try
      {
        ((java.io.RandomAccessFile)f.id).writeByte( b );
        f.res = Done;
      }
      catch (java.io.IOException e)
      {
        f.res = NotDone;
      }
    }
    else
    {
      f.res = ParamError;
    }
  );
END WriteByte;



PROCEDURE WriteChar( VAR f : File; ch : CHAR );
BEGIN
  SYSTEM.INLINE
  (
    if (f.id instanceof java.io.InputStreamReader)
    {
      f.res = ParamError;
    }
    else if (f.id instanceof java.io.OutputStreamWriter)
    {
      try
      {
        java.io.OutputStreamWriter id = (java.io.OutputStreamWriter)f.id;
        id.write( (int)ch );
        f.res = Done;
      }
      catch (java.io.IOException e)
      {
        f.res = NotDone;
      }
    }
    else if (f.id instanceof java.io.RandomAccessFile)
    {
      try
      {
        ((java.io.RandomAccessFile)f.id).writeChar( ch );
        f.res = Done;
      }
      catch (java.io.IOException e)
      {
        f.res = NotDone;
      }
    }
    else
    {
      f.res = ParamError;
    }
  );
END WriteChar;




PROCEDURE WriteString( VAR f : File; s : ARRAY OF CHAR );
VAR
  i            : LONGCARD;
BEGIN
  i := Strings.Length( s );
  SYSTEM.INLINE
  (
    if (f.id instanceof java.io.InputStreamReader)
    {
      f.res = ParamError;
    }
    else if (f.id instanceof java.io.OutputStreamWriter)
    {
      try
      {
        java.io.OutputStreamWriter id = (java.io.OutputStreamWriter)f.id;
        id.write( s, 0, i );
        f.res = Done;
      }
      catch (java.io.IOException e)
      {
        f.res = NotDone;
      }
    }
    else if (f.id instanceof java.io.RandomAccessFile)
    {
      try
      {
        java.lang.String str = new java.lang.String( s, 0, i );
        ((java.io.RandomAccessFile)f.id).writeChars( str );
        f.res = Done;
      }
      catch (java.io.IOException e)
      {
        f.res = NotDone;
      }
    }
    else
    {
      f.res = ParamError;
    }
  );
END WriteString;



PROCEDURE WriteLn( VAR f : File );
BEGIN
  SYSTEM.INLINE
  (
    if (f.id instanceof java.io.InputStreamReader)
    {
      f.res = ParamError;
    }
    else if (f.id instanceof java.io.OutputStreamWriter)
    {
      try
      {
        java.lang.String s = java.lang.System.getProperty( "line.separator" );
        java.io.OutputStreamWriter id = (java.io.OutputStreamWriter)f.id;
        id.write( s, 0, s.length() );
        f.res = Done;
      }
      catch (java.io.IOException e)
      {
        f.res = NotDone;
      }
      catch (java.lang.SecurityException e)
      {
        f.res = NotDone;
      }
    }
    else if (f.id instanceof java.io.RandomAccessFile)
    {
      try
      {
        java.lang.String s = java.lang.System.getProperty( "line.separator" );
        java.io.RandomAccessFile id = (java.io.RandomAccessFile)f.id;
        id.writeBytes( s );
        f.res = Done;
      }
      catch (java.io.IOException e)
      {
        f.res = NotDone;
      }
      catch (java.lang.SecurityException e)
      {
        f.res = NotDone;
      }
    }
    else
    {
      f.res = ParamError;
    }
  );
END WriteLn;



BEGIN
  tempcount := 0;
  Done            := done;
  NotDone         := notdone;
  LockError       := lockerror;
  PermissionError := permissionerror;
  CallError       := callerror;
  UnknownMedium   := unknownmedium;
  UnknownFile     := unknownfile;
  FilenameError   := filenameerror;
  TooManyFiles    := toomanyfiles;
  MediumFull      := mediumfull;
  ParamError      := paramerror;
END FileSystem.

