IMPLEMENTATION MODULE InOut;

(**************************************************************************
   JAVA  Modula-2 standard utility for stream input/output.

   Copyright (c) 1998 by Mill Hill & Canterbury Corporation, Ltd.
**************************************************************************)

IMPORT SYSTEM, Strings, FileSystem, Conversions;


CONST
  CR        = CHR(13);
  LF        = CHR(10);
  BS        = CHR(8);

VAR
  stdin     : SYSTEM.ADDRESS;
  stdout    : SYSTEM.ADDRESS;


PROCEDURE RedirectInput( from : ARRAY OF CHAR );
VAR
  i : LONGINT;
BEGIN
  i := Strings.Length( from );
  IF i <= 0 THEN
    Done := FALSE;
    RETURN;
  END;
  SYSTEM.INLINE
  (
    try
    {
      java.lang.String s = new java.lang.String( from, 0, i );
      in.id = new java.io.FileReader( s );
      Done = true;
    }
    catch (java.io.FileNotFoundException e)
    {
      Done = false;
      in.id = stdin;
    }
  );
  IF Done THEN
    in.tmp := FALSE;
    in.eof := FALSE;
    Strings.Assign( from, in.name );
  END;
END RedirectInput;



PROCEDURE RedirectOutput( to : ARRAY OF CHAR );
VAR
  i : LONGINT;
BEGIN
  i := Strings.Length( to );
  IF i <= 0 THEN
    Done := FALSE;
    RETURN;
  END;
  SYSTEM.INLINE
  (
    try
    {
      java.lang.String s = new java.lang.String( to, 0, i );
      out.id = new java.io.FileWriter( s );
      Done = true;
    }
    catch (java.lang.Exception e)
    {
      Done = false;
      out.id = stdout;
    }
  );
  IF Done THEN
    out.tmp := FALSE;
    out.eof := FALSE;
    Strings.Assign( to, out.name );
  END;
END RedirectOutput;



PROCEDURE OpenInput( defext : ARRAY OF CHAR );
VAR
  FileName    : ARRAY [0..127] OF CHAR;
  i,j,k       : LONGINT;
BEGIN
  ReadString( FileName );
  IF Done THEN
    i := Strings.Length( FileName );
    IF i > 0 THEN
      IF FileName[ i-1 ] = '.' THEN
        j := 0;
        k := Strings.Length( defext );
        WHILE (j < k) AND (i+j < 127-1) DO
          FileName[ i+j ] := defext[ j ];
          INC( j );
        END;
        FileName[ i+j ] := 0C;
      END;
      RedirectInput( FileName );
    ELSE
      Done := FALSE;
    END;
  END;
END OpenInput;




PROCEDURE OpenOutput( defext : ARRAY OF CHAR );
VAR
  FileName    : ARRAY [0..127] OF CHAR;
  i,j,k       : LONGCARD;
BEGIN
  ReadString( FileName );
  IF Done THEN
    i := Strings.Length( FileName );
    IF i > 0 THEN
      IF FileName[ i-1 ] = '.' THEN
        j := 0;
        k := Strings.Length( defext );
        WHILE (j < k) AND (i+j < 127-1) DO
          FileName[ i+j ] := defext[ j ];
          INC( j );
        END;
        FileName[ i+j ] := 0C;
      END;
      RedirectOutput( FileName );
    ELSE
      Done := FALSE;
    END;
  END;
END OpenOutput;



PROCEDURE CloseInput();
BEGIN
  SYSTEM.INLINE
  (
    if (in.id != stdin)
    {
      try
      {
        ((java.io.InputStreamReader)in.id).close();
      }
      catch (java.io.IOException e)
      {
      }
      in.id = stdin;
      in.res = 0;
      in.eof = false;
      in.tmp = false;
      in.name[0] = (char)0;
      Done = true;
    }
  );
END CloseInput;




PROCEDURE CloseOutput();
BEGIN
  SYSTEM.INLINE
  (
    if (out.id != stdout)
    {
      try
      {
        ((java.io.FileWriter)out.id).close();
      }
      catch (java.lang.Exception e)
      {
      }
      out.id = stdout;
      out.res = 0;
      out.eof = false;
      out.tmp = false;
      out.name[0] = (char)0;
      Done = true;
    }
  );
END CloseOutput;




PROCEDURE Read( VAR ch : CHAR );
BEGIN
  FileSystem.ReadChar( in, ch );
  Done := (in.res = FileSystem.done) AND NOT in.eof;
  termCH := ch;
END Read;




PROCEDURE ReadString( VAR s : ARRAY OF CHAR );
VAR
  i  : LONGINT;
  ch : CHAR;
BEGIN
  i := 0;
  Done := TRUE;
  ch := ' ';
  WHILE Done AND (ch = ' ') DO
    Read( ch );
  END;
  WHILE ch > ' ' DO
    IF ch = BS THEN
      DEC( i );
      IF i < 0 THEN
        i := 0;
      END;
    ELSIF ch > ' ' THEN
      IF i < VAL( LONGINT, HIGH(s) ) THEN
        s[i] := ch;
      END;
      INC(i);
    ELSE
      ch := CHR(0);
    END;
    Read( ch );
  END;
  IF i <= VAL( LONGINT, HIGH(s) ) THEN
    s[i] := 0C;
  ELSE
    s[HIGH(s)] := 0C;
  END;
  IF termCH = CR THEN
    Read( termCH );
  END;
END ReadString;




PROCEDURE ReadInt( VAR x : INTEGER );
VAR
  Buffer : ARRAY [0..255] OF CHAR;
BEGIN
  ReadString( Buffer );
  IF Done THEN
    Conversions.StrToInt( Buffer, x, Done );
  END;
END ReadInt;



PROCEDURE ReadShortInt( VAR x : SHORTINT );
VAR
  Buffer : ARRAY [0..255] OF CHAR;
BEGIN
  ReadString( Buffer );
  IF Done THEN
    Conversions.StrToShortInt( Buffer, x, Done );
  END;
END ReadShortInt;



PROCEDURE ReadLongInt( VAR x : LONGINT );
VAR
  Buffer : ARRAY [0..255] OF CHAR;
BEGIN
  ReadString( Buffer );
  IF Done THEN
    Conversions.StrToLongInt( Buffer, x, Done );
  END;
END ReadLongInt;



PROCEDURE ReadCard( VAR x : CARDINAL );
VAR
  Buffer : ARRAY [0..255] OF CHAR;
BEGIN
  ReadString( Buffer );
  IF Done THEN
    Conversions.StrToCard( Buffer, x, Done );
  END;
END ReadCard;




PROCEDURE ReadShortCard( VAR x : SHORTCARD );
VAR
  Buffer : ARRAY [0..255] OF CHAR;
BEGIN
  ReadString( Buffer );
  IF Done THEN
    Conversions.StrToShortCard( Buffer, x, Done );
  END;
END ReadShortCard;




PROCEDURE ReadLongCard( VAR x : LONGCARD );
VAR
  Buffer : ARRAY [0..255] OF CHAR;
BEGIN
  ReadString( Buffer );
  IF Done THEN
    Conversions.StrToLongCard( Buffer, x, Done );
  END;
END ReadLongCard;




PROCEDURE ReadWrd( VAR w : SYSTEM.WORD );
BEGIN
  ReadWord( w );
END ReadWrd;




PROCEDURE ReadByte( VAR b : SYSTEM.BYTE );
VAR
  ch : CHAR;
BEGIN
  ch := CHR( ORD( b ) );
  Read( ch );
  b := SHORT( ORD( ch ) );
END ReadByte;




PROCEDURE ReadWord( VAR w : SYSTEM.WORD );
VAR
  i : CARDINAL;
  lo : SHORTCARD;
  hi : SHORTCARD;
BEGIN
  i := 0;
  ReadByte( hi );
  IF Done THEN
    ReadByte( lo );
    i := LONG( hi ) * 256 + LONG( lo );
  END;
  w := i;
END ReadWord;




PROCEDURE ReadLongWord( VAR lw : SYSTEM.LONGWORD );
VAR
  i  : LONGCARD;
  lo : CARDINAL;
  hi : CARDINAL;
BEGIN
  i := 0;
  ReadWord( hi );
  IF Done THEN
    ReadWord( lo );
    i := LONG( hi ) * 256 + LONG( lo );
  END;
  lw := i;
END ReadLongWord;



PROCEDURE Write( ch : CHAR );
BEGIN
  FileSystem.WriteChar( out, ch );
  Done := out.res = FileSystem.done;
  IF Done THEN
    FileSystem.Doio( out );
    Done := out.res = FileSystem.done;
  END;
END Write;



PROCEDURE WriteLn();
BEGIN
  FileSystem.WriteLn( out );
  Done := out.res = FileSystem.done;
  IF Done THEN
    FileSystem.Doio( out );
    Done := out.res = FileSystem.done;
  END;
END WriteLn;



PROCEDURE WriteString( s : ARRAY OF CHAR );
BEGIN
  FileSystem.WriteString( out, s );
  Done := out.res = FileSystem.done;
  IF Done THEN
    FileSystem.Doio( out );
    Done := out.res = FileSystem.done;
  END;
END WriteString;



PROCEDURE WriteSizedString( VAR s : ARRAY OF CHAR; Size : LONGCARD );
VAR
  i : LONGCARD;
BEGIN
  Done := TRUE;
  i := Strings.Length( s );
  IF Size > i THEN
    i := Size - i;
    WHILE (i > 0) AND Done DO
      Write( ' ' );
      DEC( i );
    END;
  END;
  IF Done THEN
    WriteString( s );
  END;
END WriteSizedString;




PROCEDURE WriteInt( x : INTEGER; n : CARDINAL );
VAR
  Buffer : ARRAY [0..11] OF CHAR;
BEGIN
  Conversions.IntToStr( x, Buffer, Done );
  IF Done THEN
    WriteSizedString( Buffer, n );
  ELSE
    Done := FALSE;
  END;
END WriteInt;



PROCEDURE WriteShortInt( x : SHORTINT; n : CARDINAL );
VAR
  Buffer : ARRAY [0..11] OF CHAR;
BEGIN
  Conversions.ShortIntToStr( x, Buffer, Done );
  IF Done THEN
    WriteSizedString( Buffer, n );
  ELSE
    Done := FALSE;
  END;
END WriteShortInt;




PROCEDURE WriteLongInt( x : LONGINT; n : CARDINAL );
VAR
  Buffer : ARRAY [0..11] OF CHAR;
BEGIN
  Conversions.LongIntToStr( x, Buffer, Done );
  IF Done THEN
    WriteSizedString( Buffer, n );
  ELSE
    Done := FALSE;
  END;
END WriteLongInt;




PROCEDURE WriteCard( x : CARDINAL; n : CARDINAL );
VAR
  Buffer : ARRAY [0..11] OF CHAR;
BEGIN
  Conversions.CardToStr( x, Buffer, Done );
  IF Done THEN
    WriteSizedString( Buffer, n );
  ELSE
    Done := FALSE;
  END;
END WriteCard;




PROCEDURE WriteShortCard( x : SHORTCARD; n : CARDINAL );
VAR
  Buffer : ARRAY [0..11] OF CHAR;
BEGIN
  Conversions.ShortCardToStr( x, Buffer, Done );
  IF Done THEN
    WriteSizedString( Buffer, n );
  ELSE
    Done := FALSE;
  END;
END WriteShortCard;




PROCEDURE WriteLongCard( x : LONGCARD; n : CARDINAL );
VAR
  Buffer : ARRAY [0..11] OF CHAR;
BEGIN
  Conversions.LongCardToStr( x, Buffer, Done );
  IF Done THEN
    WriteSizedString( Buffer, n );
  ELSE
    Done := FALSE;
  END;
END WriteLongCard;



PROCEDURE WriteOct( x : CARDINAL; n : CARDINAL );
VAR
  Buffer : ARRAY [0..13] OF CHAR;
BEGIN
  Conversions.OctToStr( x, Buffer, Done );
  IF Done THEN
    WriteSizedString( Buffer, n );
  ELSE
    Done := FALSE;
  END;
END WriteOct;



PROCEDURE WriteShortOct( x : SHORTCARD; n : CARDINAL );
VAR
  Buffer : ARRAY [0..13] OF CHAR;
BEGIN
  Conversions.ShortOctToStr( x, Buffer, Done );
  IF Done THEN
    WriteSizedString( Buffer, n );
  ELSE
    Done := FALSE;
  END;
END WriteShortOct;



PROCEDURE WriteLongOct( x : LONGCARD; n : CARDINAL );
VAR
  Buffer : ARRAY [0..13] OF CHAR;
BEGIN
  Conversions.LongOctToStr( x, Buffer, Done );
  IF Done THEN
    WriteSizedString( Buffer, n );
  ELSE
    Done := FALSE;
  END;
END WriteLongOct;



PROCEDURE WriteHex( x : CARDINAL; n : CARDINAL );
VAR
  Buffer : ARRAY [0..10] OF CHAR;
BEGIN
  Conversions.HexToStr( x, Buffer,Done );
  IF Done THEN
    WriteSizedString( Buffer, n );
  ELSE
    Done := FALSE;
  END;
END WriteHex;



PROCEDURE WriteShortHex( x : SHORTCARD; n : CARDINAL );
VAR
  Buffer : ARRAY [0..10] OF CHAR;
BEGIN
  Conversions.ShortHexToStr( x, Buffer, Done );
  IF Done THEN
    WriteSizedString( Buffer, n );
  ELSE
    Done := FALSE;
  END;
END WriteShortHex;



PROCEDURE WriteLongHex( x : LONGCARD; n : CARDINAL );
VAR
  Buffer : ARRAY [0..10] OF CHAR;
BEGIN
  Conversions.LongHexToStr( x, Buffer, Done );
  IF Done THEN
    WriteSizedString( Buffer, n );
  ELSE
    Done := FALSE;
  END;
END WriteLongHex;



PROCEDURE WriteWrd( w : SYSTEM.WORD );
BEGIN
  WriteWord( w );
END WriteWrd;



PROCEDURE WriteWord( w : SYSTEM.WORD );
VAR
  val : CARDINAL;
BEGIN
  val := w;
  WriteByte( SHORT( val DIV 256 ) );
  IF Done THEN
    WriteByte( SHORT( val ) );
  END;
END WriteWord;



PROCEDURE WriteByte( b : SYSTEM.BYTE );
BEGIN
  FileSystem.WriteByte( out, b );
  Done := out.res = FileSystem.done;
  IF Done THEN
    FileSystem.Doio( out );
    Done := out.res = FileSystem.done;
  END;
END WriteByte;




PROCEDURE WriteLongWord( lw : SYSTEM.LONGWORD );
VAR
  val : LONGCARD;
BEGIN
  val := lw;
  WriteWord( SHORT( val DIV 10000H )  );
  IF Done THEN
    WriteWord( SHORT( val ) );
  END;
END WriteLongWord;



BEGIN
  SYSTEM.INLINE
  (
    in.id = new java.io.InputStreamReader( java.lang.System.in );
    out.id = new java.io.OutputStreamWriter( java.lang.System.out );
  );
  stdin := in.id;
  stdout := out.id;
  termCH := ' ';
END InOut.

