MODULE Slides;

(* A simple Web page click animation entirely done       *)
(* in Canterbury Modula-2 for Java.                      *)
(* Last update: Jan 17, 2000                             *)
(*                                                       *)
(* (c) 1998 by Mill Hill & Canterbury Corporation, Ltd.  *)

(*$XL+ Object Oriented Language Extensions *)
(*$XF+ Extended function designator usage  *)

IMPORT
  SYSTEM,
  java_applet_Applet,
  java_lang_Runnable,
  java_lang_Thread,
  java_awt_Image,
  java_awt_Graphics,
  java_awt_Event,
  java_lang_String,
  java_lang_RuntimeException;

TYPE
  Applet           = java_applet_Applet.Applet;
  Runnable         = java_lang_Runnable.Runnable;
  PThread          = java_lang_Thread.POINTER_TO_Thread;
  PImage           = java_awt_Image.POINTER_TO_Image;
  PString          = java_lang_String.POINTER_TO_String;
  PGraphics        = java_awt_Graphics.POINTER_TO_Graphics;
  PEvent           = java_awt_Event.POINTER_TO_Event;
  RuntimeException = java_lang_RuntimeException.RuntimeException;

  ClickAnimation = RECORD( Applet, Runnable )
    animationThread : PThread;
    cur : INTEGER;
    frame : ARRAY [0..3] OF PImage;
    mouseClick : BOOLEAN;
  END;



(* Modula-2 always regards constructors as type-bound
   procedures, but Java needs to know that it is a
   public constructor, hence the need for using the
   JAVA PUBLIC CONSTRUCTOR directive.
*)

(*$JAVA PUBLIC CONSTRUCTOR *)
PROCEDURE( VAR Self : ClickAnimation) _init_;
BEGIN
  Self._init_^;
END _init_;



(* The public type-bound procedure "init" overrides the
   one inherited from Java class java.applet.Applet.
   This procedure always gets called first in
   a Java applet and is responsible for intializing
   all the fields
*)

(*$JAVA PUBLIC *)
PROCEDURE( VAR Self : ClickAnimation) init();
  VAR
    i          : INTEGER;
    gifName    : ARRAY [0..8] OF CHAR;

  PROCEDURE JavaString( s : ARRAY OF CHAR ) : PString;
    VAR Result : PString;
    VAR i : LONGINT;
  BEGIN
    NEW( Result, SYSTEM.ADR(s) );
    i := Result^.indexOf( 0 );
    IF i >= 0 THEN
      Result := Result^.substring( 0, i );
    END;
    RETURN Result;
  END JavaString;

BEGIN
  (* set display size for applet *)
  Self.resize(300,300);

  (* get the GIF-formatted images *)
  FOR i := 0 TO 3 DO
    CASE i OF
    | 0 :        gifName := 'col1.gif';
    | 1 :        gifName := 'col2.gif';
    | 2 :        gifName := 'col3.gif';
    ELSE         gifName := 'col4.gif';
    END;
    Self.frame[i] := Self.getImage( Self.getDocumentBase(), JavaString(gifName) );
  END;

  (* no mouse click yet *)
  Self.mouseClick := FALSE;
END init;



(* The public type-bound procedure "start"
   overrides java.applet.Applet.start.
   It is called by the browser or applet viewer to inform this applet
   that it should start its execution.
   It is called after the init method and
   each time the applet is revisited in a Web page.
*)

(*$JAVA PUBLIC *)
PROCEDURE( VAR Self : ClickAnimation) start();
BEGIN
  (* Create a new thread for class ClickAnimation.
     Canterbury uses an extended standard procedure
     NEW, with an optional second parameter for
     a constrcutor call
  *)
  NEW( Self.animationThread, SYSTEM.ADR(Self) );

  (* The start() method spawns the animation thread in order
     for the animation to run the first time
  *)
  Self.animationThread^.start();

  (* Once the animation thread is started, it calls
     the run() method.
  *)
END start;



(* The type-bound procedure "stop" overrides java.applet.Applet.stop.
   It is called by the browser or applet viewer to inform this applet
   that it should stop its execution.
   It is called when the Web page that contains this applet has been
   replaced by another page, and also just before the applet is to be
   destroyed.
*)

(*$JAVA PUBLIC *)
PROCEDURE( VAR Self : ClickAnimation) stop();
BEGIN
  SYSTEM.TRY();
  Self.animationThread^.stop();
  SYSTEM.CATCH( RuntimeException );
  (* Just catch it.
     Some JVMs, especially in Java 2,
     do a security exception here.
  *)
  SYSTEM.ENDTRY();
END stop;




(* The type-bound procedure "run" implements the one from the
   Java interface java.lang.Runnable.
   Starting the thread causes the objects "run" method to be
   called in that separately executing thread.
*)

(*$JAVA PUBLIC *)
PROCEDURE( VAR Self : ClickAnimation) run();
BEGIN
  (* If there was no mouse click repaint the first frame,
     else repaint the next one.
  *)
  IF (NOT Self.mouseClick) THEN
    Self.cur := 0;
    Self.repaint();
  ELSE
    Self.mouseClick := FALSE;
    Self.cur := Self.cur + 1;
    IF (Self.cur > 3) THEN
      Self.cur := 0;
    END;
    Self.repaint();
  END;
END run;




(* The type-bound procedure "paint" overrides an inherited one.
   It is responsible for painting the applets display area.
*)

(*$JAVA PUBLIC *)
PROCEDURE( VAR Self : ClickAnimation) paint( g : PGraphics );
BEGIN
  g^.drawImage( Self.frame[Self.cur], 0, 0, SYSTEM.ADR(Self) );
END paint;



(* Method "mouseDown" overrides an inherited one.
   Always called in case of a mouse button being pressed
   while its cursor is on the applet.
*)

(*$JAVA PUBLIC *)
PROCEDURE( VAR Self : ClickAnimation) mouseDown
(
  e : PEvent;
  x : LONGINT;
  y : LONGINT
)   : BOOLEAN;
BEGIN
  Self.mouseClick := TRUE;
  NEW( Self.animationThread, SYSTEM.ADR(Self) );
  Self.animationThread^.start();
  RETURN TRUE;
END mouseDown;

END Slides.


